﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;

using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.Logging;

namespace Platform.Logging.FileLogger
{
    [ProviderAlias("File")]
    public class PlatformFileLoggerProvider : PlatformBatchLoggerProvider<PlatformFileLoggerOptions>
    {
        public PlatformFileLoggerProvider( IConfigurationSection configuration )
            : base( configuration )
        {
            _Current = Directory.GetCurrentDirectory();
        }

        protected override string SectionName => "File";

        private string _Current;
        private string _Path;
        private string _Name;

        internal override PlatformBatchLogger<PlatformFileLoggerOptions> CreateLoggerImplementation( string categoryName )
        {
            return new PlatformFileLogger(this, categoryName);
        }

        protected override void OnConfigurationChanged()
        {
            // call base class
            base.OnConfigurationChanged();

            _Path = ProviderOptions.Path ?? LoggingOptions.Path ?? _Current;
            _Name = ProviderOptions.Filename ?? "Logging";

            // create directory
            try
            {
                if (!string.IsNullOrEmpty(_Path))
                    if (!Directory.Exists(_Path))
                        Directory.CreateDirectory(_Path);
            }
            catch( Exception ex )
            {
                SimpleFileLogger.WriteLine(Configuration, "{0} Logger Failed to create directory {1}: {2}", SectionName, _Path, ex.Message);
            }
        }

        protected async override Task WriteMessagesAsync( IEnumerable<PlatformBatchLoggerMessage> messages, CancellationToken cancellationToken)
        {
            // get file size limit
            var limit = ProviderOptions.FileSizeLimit ?? 10 * 1024 * 1024;

            // get file name and path
            string filename = string.Format("{0}_{1:yyyy-MM-dd}.log", _Name, DateTime.Now);
            string fullname = Path.Combine(_Path, filename);

            try
            {
                // roll over
                var info = new FileInfo(fullname);
                for ( int index=0; info.Exists && info.Length >= limit; index++ )
                {
                    filename = string.Format("{0}_{1:yyyy-MM-dd}_{2:0000}.log", _Name, DateTime.Now, index + 1);
                    fullname = Path.Combine(_Path, filename);
                    info = new FileInfo(fullname);
                }

                // create file writer
                using (var filewriter = info.AppendText())
                {
                    // write all messages
                    foreach (var message in messages)
                    {
                        // check if canceled
                        if (cancellationToken.IsCancellationRequested)
                            return;

                        var text = message.Message.Replace("\r\n", "\n").Replace("\r", "\n").Replace("\n", "\r\n     ").Trim('\r', '\n');
                        var line1 = string.Format("{0} {1:HH:mm:ss.fffff} {2} {3} {4}", GetLogLevelString(message.LogLevel), message.Timestamp, message.Category, message.EventId.Id, message.EventId.Name);
                        var line2 = "     " + text;

                        await filewriter.WriteLineAsync(line1);
                        await filewriter.WriteLineAsync(line2);

                        if (message.Exception != null)
                            await filewriter.WriteLineAsync("     " + message.Exception.ToString().Replace("\r\n", "\r\n     "));
                    }

                    // close file writer
                    filewriter.Close();
                }
            }
            catch ( Exception ex )
            {
                SimpleFileLogger.WriteLine(Configuration, "{0} Logger Failed to write messages to {1}: {2}", SectionName, fullname, ex.Message);
            }
        }

        protected override Task CleanupLoggingAsync(int retention)
        {
            // get file name and path
            string filename = string.Format("{0}_{1:yyyy-MM-dd}.log", _Name, DateTime.Now - TimeSpan.FromDays(retention) );

            // cleanup files
            foreach( var file in Directory.GetFiles( _Path, "*.log" ).Where( f => string.Compare( Path.GetFileName(f), filename, true ) < 0 ) )
            {
                try
                {
                    File.Delete(file);
                }
                catch( Exception ex )
                {
                    SimpleFileLogger.WriteLine(Configuration, "{0} Logger Failed to delete file {1}: {2}", SectionName, file, ex.Message);
                }
            }

            // return completed task
            return Task.CompletedTask;
        }
    }
}
